// Physical memory allocator, for user processes,
// kernel stacks, page-table pages,
// and pipe buffers. Allocates whole 4096-byte pages.

#include "types.h"
#include "param.h"
#include "memlayout.h"
#include "spinlock.h"
#include "riscv.h"
#include "defs.h"

void freerange(void *pa_start, void *pa_end);

extern char end[]; // first address after kernel.
                   // defined by kernel.ld.

struct run {
  struct run *next;
};

struct {
  struct spinlock lock;
  struct run *freelist;
  char refcounts[(PHYSTOP - KERNBASE) >> PGSHIFT];
} kmem;

void
kinit()
{
  initlock(&kmem.lock, "kmem");
  memset(kmem.refcounts, 1, sizeof(kmem.refcounts));
  freerange(end, (void*)PHYSTOP);
}

void
freerange(void *pa_start, void *pa_end)
{
  char *p;
  p = (char*)PGROUNDUP((uint64)pa_start);
  for(; p + PGSIZE <= (char*)pa_end; p += PGSIZE)
    kfree(p);
}

int inc_refcount(void *pa, int delta, int havelock) {
  if (!havelock)
    acquire(&kmem.lock);
  uint64 index = (((uint64) pa) - KERNBASE) >> PGSHIFT;
  if (index > sizeof(kmem.refcounts))
    panic("inc_refcount called with pointer outside range");
  int count = kmem.refcounts[index] += (char) delta;
  if (!havelock)
    release(&kmem.lock);
  return count;
}

// Free the page of physical memory pointed at by pa,
// which normally should have been returned by a
// call to kalloc().  (The exception is when
// initializing the allocator; see kinit above.)
void
kfree(void *pa)
{
  struct run *r;

  if(((uint64)pa % PGSIZE) != 0 || (char*)pa < end || (uint64)pa >= PHYSTOP)
    panic("kfree");

  int count = inc_refcount(pa, -1, 0);
  if (count > 0) {
      return;
  }
  if (count < 0)
      panic("kfree ended up with negative refcount");

  // Fill with junk to catch dangling refs.
  memset(pa, 1, PGSIZE);

  r = (struct run*)pa;

  acquire(&kmem.lock);
  r->next = kmem.freelist;
  kmem.freelist = r;
  release(&kmem.lock);
}

// Allocate one 4096-byte page of physical memory.
// Returns a pointer that the kernel can use.
// Returns 0 if the memory cannot be allocated.
void *
kalloc(void)
{
  struct run *r;

  acquire(&kmem.lock);
  r = kmem.freelist;
  if(r) {
    kmem.freelist = r->next;
    int count = inc_refcount(r, 1, 1);
    if (count != 1)
        panic("kalloc got page with non-zero refcount");
  }
  release(&kmem.lock);

  if(r)
    memset((char*)r, 5, PGSIZE); // fill with junk
  return (void*)r;
}
