#include "types.h"
#include "param.h"
#include "riscv.h"
#include "defs.h"
#include "spinlock.h"
#include "proc.h"

void
print_pages(pagetable_t table, int level, int continues)
{
    int last_valid = -1;
    for (int i = 0; i < 512; i++) {
        pte_t pte = table[i];
        if (pte & PTE_V)
            last_valid = i;
    }
    int prev = last_valid + 1;

    for (int i = 511; i >= 0; i--) {
        pte_t pte = table[i];
        if (!(pte & PTE_V))
            continue;

        // print ... for gaps
        if (prev != i + 1) {
            for (int indent = 2; indent > level; indent--) {
                if (continues & (1<<indent))
                    printf("│       ");
                else
                    printf("        ");
            }
            printf("├─ ...\n");
        }
        prev = i;

        // descend to deeper levels
        if (level > 0) {
            print_pages((pagetable_t) PTE2PA(pte), level-1,
                i < last_valid ? continues | (1<<level) : continues);
        }

        // print continuation lines/indents for higher levels
        for (int indent = 2; indent > level; indent--) {
            if (continues & (1<<indent))
                printf("│       ");
            else
                printf("        ");
        }

        // print lead line to this level
        if (i == last_valid)
            printf("┌────── ");
        else
            printf("├────── ");

        // print the entry
        printf("[%d] %s─⯈ %p", i, i < 10 ? "──" : i < 100 ? "─" : "", (void *) PTE2PA(pte));
        if (level == 0) {
            if (!(pte & PTE_R))
                printf(" !r");
            if (!(pte & PTE_W))
                printf(" !w");
            if (!(pte & PTE_X))
                printf(" !x");
            if (!(pte & PTE_U))
                printf(" !u");
        }
        printf("\n");
    }

    if (level == 2)
        printf("[satp] ─⯈ %p\n", table);
}

extern struct proc proc[NPROC];

uint64
sys_pages(void)
{
  int pid;
  argint(0, &pid);
  int found = 0;

  for (struct proc *p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if (p->pid > 0) {
      if ((pid > 0 && p->pid == pid) || pid == 0) {
        found = 1;
        printf("\npage tables for process %d:\n", p->pid);
        print_pages(p->pagetable, 2, 0);
      }
    }
    release(&p->lock);
  }

  return found - 1;
}
